
import inspect
from collections import OrderedDict

from django.core.exceptions import ImproperlyConfigured
from django.utils import six
from django import forms
from django.utils.encoding import force_text
from djingles.formatters import Formatter
from djingles.forms.fields import SortField
from djingles import utils


__all__ = ['ActionFormMixin', 'ActionModelForm', 'ActionForm', 'FilterForm',
           'FilterModelForm', 'FilterFormMixin', 'action_model_form_factory']


class ActionFormMixin(object):

    confirmation_message = None
    failure_message = None
    success_message = None

    def __init__(self, **kwargs):
        self.context = kwargs.pop("context", {})
        self.action = self.context.get("action", "execute")
        self.context["action"] = self.action
        self.action_method = getattr(self, self.action)
        super(ActionFormMixin, self).__init__(**kwargs)
        prepare = getattr(self, 'prepare_%s' % self.action, None)
        if prepare is not None:
            prepare()

    def __bind_fields(self):
        key = "__bound_with_form"
        for name, field in self.fields.items():
            if hasattr(field, 'bind_form') and not getattr(field, key, False):
                setattr(field, key, True)
                field.bind_form(self)

    def process_context(self, func):
        context = self.context.copy()
        if hasattr(self, "cleaned_data") and hasattr(self, "save"):
            instance = self.save(commit=False)
            context["instance"] = instance
        context["data"] = self.cleaned_data
        spec = inspect.getfullargspec(func)
        if spec.varargs:
            raise ImproperlyConfigured("Form action cannot have variable arguments")
        if spec.varkw:
            return context
        return {k: context[k] for k in spec.args[1:] if k in context}

    def process_result(self, result):
        return result

    @property
    def result(self):
        self.is_valid()
        return self.__result

    def get_success_message(self):
        if self.success_message:
            return self.success_message.format(**self.context)
        return self.success_message

    def get_failure_message(self):
        if self.failure_message:
            return self.failure_message.format(**self.context)
        return self.failure_message

    def get_confirmation_message(self):
        if self.confirmation_message:
            return self.confirmation_message.format(**self.context)
        return self.confirmation_message

    @classmethod
    def class_oid(cls):
        """
        Obfuscated class id
        :return: str
        """
        return utils.create_hash(utils.qualified_name(cls))

    @classmethod
    def call(cls, *kwargs):
        form_obj = cls(**kwargs)
        if not form_obj.is_valid():
            return False, form_obj.errors
        else:
            return True, form_obj.result

    def full_clean(self):
        result = None
        self.__bind_fields()
        super(ActionFormMixin, self).full_clean()
        try:
            _ = self.__result
        except AttributeError:
            try:
                if self.is_bound and not self._errors:
                    method = self.action_method
                    context = self.process_context(method)
                    result = method(**context)
                    result = self.process_result(result)
            except forms.ValidationError as ex:
                self.add_error(None, ex)
            finally:
                self.__result = result

    def select_fields(self, exclude=None, include=None, **extras):
        if exclude is not None:
            for f in exclude:
                self.fields.pop(f, None)
        if include is not None:
            include = set(include)
            for f in list(self.fields.keys()):
                if f not in include:
                    self.fields.pop(f)
        if extras:
            for f in extras:
                self.fields[f] = extras[f]

    def execute(self, **kwargs):
        return {}


class ActionForm(ActionFormMixin, forms.Form):
    pass


class ActionModelForm(ActionFormMixin, forms.ModelForm):

    confirmation_message = "Please confirm that you really wish to {action} this object ?"

    def prepare_delete(self):
        self.select_fields(include=())

    def delete(self):
        self.instance.delete()

    def update(self):
        return self.save()

    def create(self):
        return self.save()

    def clone(self):
        self.instance.id = None
        return self.save()


class FilterFormMixin:

    def __init__(self, *args, **kwargs):
        self.context = kwargs.pop("context", {})
        super(FilterFormMixin, self).__init__(*args, **kwargs)
        for f in self.fields.values():
            f.required = False

    def __bind_fields(self):
        key = "__bound_with_form"
        for name, field in self.fields.items():
            if hasattr(field, 'bind_form') and not getattr(field, key, False):
                setattr(field, key, True)
                field.bind_form(self)

    @property
    def sort_field(self):
        for name, f in self.fields.items():
            if isinstance(f, SortField):
                return self[name]

    def generic_filter(self, name, queryset, value, data):
        kwargs = {}
        if isinstance(value, (tuple, list)):
            name = '%s__in' % name
        kwargs[name] = value
        return queryset.filter(**kwargs)

    def perform_filter(self, queryset):
        self.__bind_fields()
        if not self.is_valid():
            return self.invalid_queryset(queryset)
        data = self.cleaned_data
        allowed = set(self.fields.keys())
        if hasattr(self, "before_filters"):
            queryset = self.before_filters(queryset, data)
        for name, value in six.iteritems(data):
            if name not in allowed or value in (None, '', [], ()):
                continue
            kwargs = {}
            field = self.fields[name]
            if hasattr(self, "filter_%s" % name):
                result = getattr(self, "filter_%s" % name)(queryset, value, data)
            elif hasattr(field, "filter_queryset"):
                result = field.filter_queryset(queryset, value, self[name])
            else:
                result = self.generic_filter(name, queryset, value, data)
            if result is not None:
                queryset = result
        if hasattr(self, "after_filters"):
            queryset = self.after_filters(queryset, data)
        return queryset

    def invalid_queryset(self, queryset):
        return queryset.none()


class FilterForm(FilterFormMixin, forms.Form):
    pass


class FilterModelForm(FilterFormMixin, forms.ModelForm):
    pass


def action_model_form_factory(model_class, include=None, exclude=(), **kwargs):
    name = "%sActionForm" % model_class.__name__
    meta = type("Meta", (), {"include": include, "exclude": exclude, "model": model_class})
    kwargs['Meta'] = meta
    return type(name, (ActionModelForm, ), kwargs)




from django import forms
from django.contrib import messages
from django.forms.models import ModelForm
from django.http.response import Http404, HttpResponseGone
from django.shortcuts import redirect
from django.views.generic.base import TemplateResponseMixin, View
from djingles import utils, exceptions


__all__ = ['CommonView', 'CommonSessionDataMixin', 'CommonTemplateView', 'CommonFormView']



class CommonSessionDataMixin(object):

    session_key_prefix = ""

    def get_session_key(self):
        return "%s-%s" % (self.session_key_prefix, self.class_oid())

    def get_session_data(self):
        return self.request.session.get(self.get_session_key())

    def set_session_data(self, data):
        self.request.session[self.get_session_key()] = data

    def clear_session_data(self):
        self.request.session.pop(self.get_session_key(), None)

    @classmethod
    def class_oid(cls):
        return utils.create_hash(utils.qualified_name(cls))


class CommonView(View, CommonSessionDataMixin):

    DEBUG = messages.DEBUG
    INFO = messages.INFO
    WARNING = messages.WARNING
    ERROR = messages.ERROR
    SUCCESS = messages.SUCCESS

    def get_user(self):
        return self.request.user

    def process_request(self, request):
        self.user = self.get_user()

    def process_response(self, request, response):
        return response

    def process_exception(self, request, ex):
        if isinstance(ex, exceptions.Redirect):
            return redirect(ex.create_url(request), permanent=ex.permanent)

    def dispatch(self, request, *args, **kwargs):
        try:
            response = self.process_request(request)
            if not response:
                response = super(CommonView, self).dispatch(request, *args, **kwargs)
        except Exception as ex:
            response = self.process_exception(request, ex)
            if response is None:
                raise
        response = self.process_response(request, response)
        return response

    def get_context_data(self, **kwargs):
        if 'view' not in kwargs:
            kwargs['view'] = self
        return kwargs

    def add_message(self,  message, level=messages.INFO, **kwargs):
        if not message:
            return
        messages.add_message(self.request, level, message, **kwargs)

    def get_messages(self):
        return messages.get_messages(self.request)

    def get_previous_url(self, default=None):
        return self.request.META.get("HTTP_REFERER", default)


class CommonTemplateView(CommonView, TemplateResponseMixin):

    view_icon = None

    view_label = None

    page_title = None

    page_heading = None

    page_actions = ()

    page_css_class = None

    def __init__(self, **kwargs):
        super(CommonTemplateView, self).__init__(**kwargs)
        self.extra_context = {}

    def get(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)
        return self.render_to_response(context)

    def get_page_css_class(self):
        return self.page_css_class

    def get_page_title(self):
        return self.page_title or self.get_page_heading()

    def get_page_heading(self):
        return self.page_heading

    def get_page_actions(self):
        return self.page_actions

    def get_view_icon(self):
        return self.view_icon

    def get_view_label(self):
        return self.view_label

    def render_to_response(self, ctx, **response_kwargs):
        ctx.update(self.extra_context)
        if not self.request.is_ajax():
            ctx['view'] = self
            ctx.setdefault("messages", self.get_messages())
            ctx.setdefault('page_heading', self.get_page_heading())
            ctx.setdefault('page_title', self.get_page_title())
            ctx.setdefault('page_actions', self.get_page_actions())
            ctx.setdefault('page_css_class', self.get_page_css_class())
        response = super(CommonTemplateView, self).render_to_response(ctx, **response_kwargs)
        return response


class CommonFormView(CommonTemplateView):

    success_url = None
    form_class = None

    context_form_key = "form"

    def get_context_form_key(self, form):
        key = self.context_form_key or "form"
        return key

    def get_success_url(self, form):
        return self.success_url

    def get_form_class(self, form_key=None):
        return self.form_class

    def get_form_key(self):
        return

    def can_submit(self):
        return self.request.method == "POST"

    def post(self, request, *args, **kwargs):
        if self.can_submit():
            return self.process_submit(
                form_key=self.get_form_key(),
                data=request.POST,
                files=request.FILES
            )
        return redirect(request.get_full_path())

    def get(self, request, *args, **kwargs):
        form_key = self.get_form_key()
        if self.can_submit():
            return self.process_submit(form_key=form_key, data=request.GET)
        else:
            data, files = self.get_form_data(form_key)
            form = self.get_form(form_key=form_key, data=data, files=files)
            return self.render_form(form)

    def get_form_data(self, form_key):
        return None, None

    def set_form_data(self, form_key, data, files=None):
        raise NotImplementedError

    def form_valid(self, form):
        url = self.get_success_url(form)
        msg = form.get_success_message()
        self.add_message(level=messages.SUCCESS, message=msg)
        return redirect(url)

    def form_invalid(self, form):
        msg = form.get_failure_message()
        self.add_message(level=messages.ERROR, message=msg)
        return self.render_form(form)

    def render_form(self, form):
        form_key = self.get_form_key()
        kwargs = {self.get_context_form_key(form_key): form}
        return self.render_to_response(self.get_context_data(**kwargs))

    def get_form_initial(self, form_key):
        return None

    def get_form_prefix(self, form_key):
        return None

    def get_form_instance(self, form_key):
        return None

    def get_form_context(self, form_key):
        action = self.action
        ctx = {
            "request": self.request,
            "user": self.user,
        }
        if action:
            ctx['action'] = action
        return ctx

    def get_form_kwargs(self, form_key):
        kwargs = {
            "initial": self.get_form_initial(form_key),
            "prefix": self.get_form_prefix(form_key),
        }
        form_class = self.get_form_class(form_key)
        if form_class and issubclass(form_class, ModelForm):
            kwargs['instance'] = self.get_form_instance(form_key)
        kwargs['context'] = self.get_form_context(form_key)
        return kwargs

    def process_submit(self, form_key, data=None, files=None):
        form_obj = self.get_form(form_key, data=data, files=files)
        if form_obj.is_valid():
            return self.form_valid(form_obj)
        else:
            return self.form_invalid(form_obj)

    def get_form(self, form_key, data=None, files=None):
        form_class = self.get_form_class(form_key)
        kwargs = self.get_form_kwargs(form_key)
        kwargs['data'] = data
        kwargs['files'] = files
        form_obj = form_class(**kwargs)
        return form_obj


class CommonMultipleFormView(CommonFormView):

    form_classes = {}
    form_key_parameter = '_fkey'

    def get_context_form_key(self, form_key):
        suffix = "form"
        return suffix if not form_key else "%s_%s" % (form_key, suffix)

    def get_form_key(self):
        if not self.can_submit():
            return None
        name = self.form_key_parameter
        request = self.request
        data = request.GET if request.method == 'GET' else request.POST
        for key in self.form_classes:
            field_name = '%s-%s' % (key, name)
            if field_name in data:
                return data[field_name]

    def get_form(self, form_key, data=None, files=None):
        form_obj = super(CommonMultipleFormView, self).get_form(form_key, data, files)
        form_obj.fields[self.form_key_parameter] = forms.CharField(widget=forms.HiddenInput, initial=form_key)
        return form_obj

    def get_form_prefix(self, form_key):
        return form_key

    def get_form_class(self, form_key):
        return self.form_classes[form_key]

    def get_unbound_form_keys(self):
        return [k for k in self.form_classes.keys() if k != self.get_form_key()]

    def get(self, request, *args, **kwargs):
        key = self.get_form_key()
        if key is None:
            context = self.get_context_data(**kwargs)
            return self.render_to_response(context)
        return self.process_submit(key, data=request.GET, files=None)

    def post(self, request, *args, **kwargs):
        key = self.get_form_key()
        if key is None:
            return redirect(request.get_full_path())
        return self.process_submit(key, data=request.POST, files=request.FILES)

    def get_context_data(self, **kwargs):
        context = super(CommonMultipleFormView, self).get_context_data(**kwargs)
        for form_key in self.get_unbound_form_keys():
            data, files = self.get_form_data(form_key)
            form = self.get_form(form_key=form_key, data=data, files=files)
            context[self.get_context_form_key(form_key)] = form
        return context




import inspect
import functools
from django.urls import reverse, path
from django.http.response import Http404, HttpResponse, HttpResponseNotAllowed
from django.shortcuts import redirect
from django.urls.exceptions import NoReverseMatch

from djingles.formatters.models import object_formatter_factory, table_formatter_factory
from djingles.forms import action_model_form_factory
from djingles.views.generic import CommonFormView, CommonTemplateView

from djingles.pagination import Paginator


__all__ = [
           'CommonViewSet',
           'CreateViewSetMixin',
           'CommonModelViewSet',
           'UpdateViewSetMixin',
           'ListViewSetMixin',
           'DeleteViewSetMixin',
           'DetailViewSetMixin',
            'CloneViewSetMixin',
            'detail_view',
            'list_view'
           ]


class _SubView(object):
    detail = False
    regex = None
    methods = None

    def __init__(self, name, **kwargs):
        for k,v in kwargs.items():
            setattr(self, k, v)
        self.name = name


def view(**kwargs):
    methods = kwargs.pop("methods", None)

    def wrapper(func):
        subview = _SubView(name=func.__name__, **kwargs)
        wrapped = func

        if methods:
            def closure(self, request, *args, **kw):
                if methods and request.method not in methods:
                    return HttpResponseNotAllowed(request.method)
                response = func(self, request, *args, **kw)
                return response
            wrapped = functools.update_wrapper(closure, func)

        wrapped.__subview__ = subview
        return func

    return wrapper


def list_view(**kwargs):
    kwargs['detail'] = False
    return view(**kwargs)


def detail_view(**kwargs):
    kwargs['detail'] = True
    return view(**kwargs)


def get_child_views(cls):
    for name, func in inspect.getmembers(cls):
        if callable(func) and hasattr(func, '__subview__'):
            yield func.__subview__


class ViewSetMixin:

    action = None
    base_name = None
    url_name = None
    url_regex = None

    def check_object_permissions(self, obj):
        pass

    def check_permissions(self):
        pass

    def get_template_names(self):
        template_name = self.template_name
        return [template_name.format(action=self.action)]

    def reverse(self, action_name, **kwargs):
        url_name = "%s_%s" % (self.base_name, action_name)
        return reverse(url_name, kwargs=kwargs)

    def get(self, request, *args, **kwargs):
        self.check_permissions()
        return getattr(self, self.action)(request)

    def post(self, request, *args, **kwargs):
        self.check_permissions()
        return getattr(self, self.action)(request)

    def process_response(self, request, response):
        if isinstance(response, dict):
            response = self.render_to_response(self.get_context_data(**response))
        return response

    @classmethod
    def as_urls(cls, base_name, url_name=None, **kwargs):
        result = []
        if url_name is None:
            url_name = base_name
        for subview in get_child_views(cls):
            action = subview.name
            segments = [url_name]
            if subview.detail:
                segments.append("<int:object_id>")
            if subview.regex is not None:
                regex = subview.regex
            else:
                regex = action
            if regex:
                segments.append(regex)
            segments.append("")
            view_url_name = "%s_%s" % (base_name, action)
            url_regex = "/".join(segments)
            view_class = path(url_regex, cls.as_view(
                base_name=base_name,
                action=subview.name,
                url_name=view_url_name,
                url_regex=url_regex,
                **kwargs
            ), name=view_url_name)
            result.append(view_class)
        return result


class DetailViewSetMixin(object):

    @view(detail=True, regex="")
    def detail(self, request):
        self.object = self.get_object()
        ctx = {
            self.context_object_key: self.object
        }
        object_formatter = self.get_object_formatter()
        if object_formatter:
            ctx[self.context_formatted_object_key] = object_formatter(self.object, variant="detail")
        context = self.get_context_data(**ctx)
        return self.render_to_response(context)


class CreateViewSetMixin(object):

    @view()
    def create(self, request):
        return self.handle_object_form()


class UpdateViewSetMixin(object):

    @view(detail=True)
    def update(self, request):
        self.object = self.get_object()
        return self.handle_object_form()


class DeleteViewSetMixin(object):

    @view(detail=True)
    def delete(self, request):
        self.object = self.get_object()
        return self.handle_object_form()


class CloneViewSetMixin(object):

    @view(detail=True)
    def clone(self, request):
        self.object = self.get_object()
        self.object.id = None
        self.success_url = self.reverse("list")
        return self.handle_object_action()


class ListViewSetMixin(object):

    context_page_key = 'object_list_page'
    context_order_key = 'order'

    @view(regex="")
    def list(self, request):

        object_list_formatter = self.get_object_list_formatter()

        object_list = self.filter_queryset(self.get_queryset(), formatter_class=object_list_formatter)

        filter_form = self.filter_form

        ctx = {}

        object_list = self.paginate_queryset(object_list)

        ctx[self.context_page_key] = object_list

        if object_list_formatter:
            object_list = object_list_formatter(object_list,
                                                sort_key=self.context_order_key,
                                                sort_field=filter_form.fields.get(self.context_order_key) if filter_form else None)

        ctx[self.context_object_list_key] = object_list

        ctx['filter_form'] = self.filter_form

        return self.render_to_response(self.get_context_data(**ctx))


class CommonViewSet(ViewSetMixin, CommonTemplateView):
    pass


class CommonModelViewSet(ViewSetMixin, CommonFormView):

    filter_class = None

    object_formatter = None
    object_list_formatter = None
    action_fields = None

    paginator = Paginator
    url_object_key = 'object_id'
    context_object_key = 'object'
    context_object_list_key = 'object_list'
    context_formatted_object_key = 'formatted_object'
    object_formatter_fields = None
    object_list_formatter_fields = None
    params_page_key = 'page'
    per_page = None

    OK_BACK = 1
    YES_BACK = 2
    CONFIRM_BACK = 3
    SUBMIT_BACK = 4

    def __init__(self, *args, **kwargs):
        super(CommonModelViewSet, self).__init__(*args, **kwargs)
        self.extra_context = {}
        self.form_context = {}
        self.form_initial = {}
        self.form_instance = None
        self.filter_form = None

    def render_to_response(self, ctx, **response_kwargs):
        self.extra_context.update(ctx)
        return super(CommonModelViewSet, self).render_to_response(self.extra_context, **response_kwargs)

    def get_queryset(self):
        return self.queryset

    def filter_queryset(self, queryset, **kwargs):
        self.filter_form = self.get_filter_form(**kwargs)
        if self.filter_form:
            queryset = self.filter_form.perform_filter(queryset)
        return queryset

    def get_filter_kwargs(self):
        return {
            "initial": self.get_filter_initial(),
            "data": self.request.GET,
            "files": None,
            "context": {"request": self.request, "user": self.user}
        }

    def get_object_formatter(self):
        if self.object_formatter is None:
            model_class = self.get_queryset().model
            self.object_formatter = object_formatter_factory(
                model_class,
                fields=self.object_formatter_fields
            )
        return self.object_formatter

    def get_object_list_formatter(self):
        if self.object_list_formatter is None:
            model_class = self.get_queryset().model
            self.object_list_formatter = table_formatter_factory(
                model_class,
                fields=self.object_list_formatter_fields,
                get_cell_url=lambda me, cell: cell.source.get_absolute_url() if
                        hasattr(cell.source, 'get_absolute_url') else self.reverse("detail", object_id=cell.source.id)
            )
        return self.object_list_formatter

    def get_form_class(self, form_key=None):
        form_class = super(CommonModelViewSet, self).get_form_class(form_key)
        if form_class is None:
            model_class = self.get_queryset().model
            form_class = action_model_form_factory(model_class, include=self.action_fields)
        return form_class

    def get_filter_initial(self):
        return None

    def get_filter_form(self, **extra):
        filter_class = self.get_filter_class()
        if filter_class:
            kwargs = self.get_filter_kwargs()
            kwargs["context"].update(extra)
            return filter_class(**kwargs)

    def paginate_queryset(self, queryset):
        if not self.per_page:
            return queryset
        return self.paginator(
            queryset,
            per_page=self.per_page,
            parameter_name=self.params_page_key,
            allow_empty=False
        ).page(self.request)

    def get_object(self):
        queryset = self.filter_queryset(self.get_queryset())
        try:
            obj = queryset.get(pk=self.kwargs[self.url_object_key])
            self.check_object_permissions(obj)
            return obj
        except queryset.model.DoesNotExist:
            raise Http404

    def get_form_action(self):
        return self.action

    def get_form_context(self, form_key):
        ctx = super(CommonModelViewSet, self).get_form_context(form_key)
        ctx['request'] = self.request
        ctx['action'] = self.get_form_action()
        ctx.update(self.form_context)
        return ctx

    def get_form_initial(self, form_key):
        initial = super(CommonModelViewSet, self).get_form_initial(form_key) or {}
        initial.update(self.form_initial)
        return initial

    def get_form_instance(self, form_key):
        form_instance = self.form_instance
        return form_instance if form_instance is not None else getattr(self, self.context_object_key, None)

    def get_filter_class(self):
        return self.filter_class

    def render_context(self, context):
        return self.render_to_response(self.get_context_data(**context))

    def handle_object_action(self, **kwargs):
        if not hasattr(self, 'object'):
            self.object = self.get_object()
        return self.handle_object_form(**kwargs)

    def handle_object_form(self, **kwargs):
        request = self.request
        method = request.method
        self.extra_context.update(kwargs)
        if method == 'GET':
            form = self.get_form(None, data=None, files=None)
            return self.render_form(form)
        else:
            try:
                if self.success_url is None:
                    if self.action in {'delete', "create"}:
                        self.success_url = self.reverse('list')
                    else:
                        self.success_url = self.reverse('detail', object_id=self.object.id)
            except NoReverseMatch:
                pass
            return self.process_submit(None, data=request.POST, files=request.FILES)

    def handle_queryset_action(self):
        request = self.request
        method = request.method
        if method == 'GET':
            return HttpResponse("Bad Request", status=400)
        else:
            object_list = self.get_queryset()
            form = self.filter_class(queryset=object_list, action=self.action, data=request.GET)
            if form.is_valid():
                return redirect(self.get_success_url(form))
            else:
                return redirect(self.get_previous_url())


